/*
 * Copyright (c) 2001-2003 Swedish Institute of Computer Science.
 * All rights reserved.
 * Copyright 2016 IS2T. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
 * SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * This file is part of the lwIP TCP/IP stack.
 *
 * Author: Adam Dunkels <adam@sics.se>
 *
 */

#include "lwip/opt.h"
#include "lwip/tcpip.h"
#include "lwip/timers.h"
#include "netif/etharp.h"
#include "ethernetif.h"
#include <string.h>
#include "LLNET_Cfg.h"
#include "LLNET_Util.h"

#include "device.h"
#include "fsl_enet.h"
#include "fsl_phy.h"
#include "fsl_sim.h"

#include "FreeRTOS.h"
#include "semphr.h"
#include "task.h"

#include "LLNET_Cfg.h"
#include "LLNET_Util.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////
/* Stack size of the interface thread */
#define INTERFACE_THREAD_STACK_SIZE            	( 350 )
#define INTERFACE_THREAD_PRIORITY              	( 14 )
/* Time waiting to check new interface link status */
#define INTERFACE_LINK_STATUS_PERIOD	500

/* MAC ADDRESS: MAC_ADDR0:MAC_ADDR1:MAC_ADDR2:MAC_ADDR3:MAC_ADDR4:MAC_ADDR5 */
#define MAC_ADDR0   0
#define MAC_ADDR1   0
#define MAC_ADDR2   0
#define MAC_ADDR3   0
#define MAC_ADDR4   0
#define MAC_ADDR5   0

/* Define those to better describe your network interface. */
#define IFNAME0 'e'
#define IFNAME1 'n'

/*******************************************************************************
 * Variables
 ******************************************************************************/
enet_handle_t g_handle;
xSemaphoreHandle enet_receive_sem = NULL;

/* MAC address buffer */
static uint8_t macaddress[6];

#if defined(__ICCARM__)
#pragma data_alignment = ENET_BUFF_ALIGNMENT
#endif
__ALIGN_BEGIN enet_rx_bd_struct_t RxBuffDescrip[ENET_RXBD_NUM] __ALIGN_END;
#if defined(__ICCARM__)
#pragma data_alignment = ENET_BUFF_ALIGNMENT
#endif
__ALIGN_BEGIN enet_tx_bd_struct_t TxBuffDescrip[ENET_TXBD_NUM] __ALIGN_END;

#if defined(__ICCARM__)
#pragma data_alignment = ENET_BUFF_ALIGNMENT
#endif
__ALIGN_BEGIN uint8_t RxDataBuff[ENET_RXBD_NUM][ENET_BuffSizeAlign(ENET_RXBUFF_SIZE)] __ALIGN_END;

#if defined(__ICCARM__)
#pragma data_alignment = ENET_BUFF_ALIGNMENT
#endif
__ALIGN_BEGIN uint8_t TxDataBuff[ENET_TXBD_NUM][ENET_BuffSizeAlign(ENET_TXBUFF_SIZE)] __ALIGN_END;

/**
 * In this function, the hardware should be initialized.
 * Called from ethernetif_init().
 *
 * @param netif the already initialized lwip network interface structure
 *        for this ethernetif
 */
void ethernetif_input(void *param);

static uint8_t* get_ethernet_mac_addr()
{
	// set MAC address to the default value
	uint8_t default_macaddress[6]= { MAC_ADDR0, MAC_ADDR1, MAC_ADDR2, MAC_ADDR3, MAC_ADDR4, MAC_ADDR5 };
	sim_uid_t deviceUID;

    // check MAC address configuration requested by MicroEJ workbench extension
	int32_t mac_conf_enabled = ((int32_t)&LLNET_MAC_CONF == LLNET_TRUE);
	if(mac_conf_enabled){
		// MAC address configuration overriden by workbench extension
		parseMacAddr(LLNET_getMACAddr(), &macaddress[0], 0, 6);

		if (macaddress[3] == 0 && macaddress[4] == 0 && macaddress[5] == 0) {
			// Consider only the prefix and generate an ID based on the mcu unique identifier
			SIM_GetUniqueId(&deviceUID);

			uint32_t crc24 = computeCRC24(&deviceUID,
					UNIQUE_DEVICE_ID_SIZE);
			macaddress[3] = crc24 & 0xFF;
			macaddress[4] = (crc24 >> 8) & 0xFF;
			macaddress[5] = (crc24 >> 16) & 0xFF;
		}
	}else{
		// set default MAC address
		memcpy(macaddress, default_macaddress, 6);
	}

	return macaddress;
}

void ethernet_callback(ENET_Type *base, enet_handle_t *handle,
		enet_event_t event, void *param) {
	switch (event) {
	case kENET_RxEvent:
	{
		portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;

		/* Give the semaphore to wakeup LwIP task */
		xSemaphoreGiveFromISR(enet_receive_sem, &xHigherPriorityTaskWoken);

		/* Switch tasks if necessary. */
		if (xHigherPriorityTaskWoken != pdFALSE) {
			portEND_SWITCHING_ISR(xHigherPriorityTaskWoken);
		}
	}
		break;
	default:
		break;
	}
}

err_t low_level_init(struct netif *netif) {
	uint8_t* macaddress = get_ethernet_mac_addr();

	/* configure ethernet peripheral (GPIOs, clocks, MAC, DMA) */
	enet_config_t config;

	err_t status = ENET_ERROR;

	uint32_t sysClock = CLOCK_GetFreq(kCLOCK_CoreSysClk);

	ENET_GetDefaultConfig(&config); // RMII 100M full duplex

	/* set netif MAC hardware address length */
	netif->hwaddr_len = ETHARP_HWADDR_LEN;

	/* set netif MAC hardware address */
	netif->hwaddr[0] = macaddress[0];
	netif->hwaddr[1] = macaddress[1];
	netif->hwaddr[2] = macaddress[2];
	netif->hwaddr[3] = macaddress[3];
	netif->hwaddr[4] = macaddress[4];
	netif->hwaddr[5] = macaddress[5];

	/* maximum transfer unit */
	netif->mtu = 1500;

	/* don't set NETIF_FLAG_ETHARP if this device is not an ethernet one */
	netif->flags = NETIF_FLAG_BROADCAST | NETIF_FLAG_ETHARP;

	if (kStatus_Success == PHY_Init(ENET, 0, sysClock)) {

		/* prepare the buffer configuration. */
		enet_buffer_config_t buffCfg = {
		ENET_RXBD_NUM,
		ENET_TXBD_NUM, ENET_BuffSizeAlign(ENET_RXBUFF_SIZE), ENET_BuffSizeAlign(
				ENET_TXBUFF_SIZE), &RxBuffDescrip[0], &TxBuffDescrip[0],
				&RxDataBuff[0][0], &TxDataBuff[0][0], };

		/* Default config */
		config.miiSpeed = kENET_MiiSpeed100M;
		config.miiDuplex = kENET_MiiFullDuplex;
		config.interrupt = kENET_RxFrameInterrupt;

		ENET_Init(ENET, &g_handle, &config, &buffCfg, netif->hwaddr, sysClock);
		ENET_SetCallback(&g_handle, ethernet_callback, netif);
		ENET_ActiveRead(ENET);

		/* create binary semaphore used for informing ethernetif of frame reception */
		if (enet_receive_sem == NULL) {
			vSemaphoreCreateBinary(enet_receive_sem);
		}

		xSemaphoreTake(enet_receive_sem, 0);

		/* create the task that handles the ETH_MAC */
		xTaskCreate((TaskFunction_t ) ethernetif_input, (const char* ) "EthIf",
				INTERFACE_THREAD_STACK_SIZE, netif, INTERFACE_THREAD_PRIORITY,
				NULL);

		status = ENET_OK;

	}

	return status;
}

/**
 * This function should do the actual transmission of the packet. The packet is
 * contained in the pbuf that is passed to the function. This pbuf
 * might be chained.
 *
 * @param netif the lwip network interface structure for this ethernetif
 * @param p the MAC packet to send (e.g. IP packet including MAC addresses and type)
 * @return ERR_OK if the packet could be sent
 *         an err_t value if the packet couldn't be sent
 *
 * @note Returning ERR_MEM here if a DMA queue of your MAC is full can lead to
 *       strange results. You might consider waiting for space in the DMA queue
 *       to become availale since the stack doesn't retry to send a packet
 *       dropped because of memory failure (except for the TCP timers).
 */

static err_t low_level_output(struct netif *netif, struct pbuf *packetBuffer)
{
    struct pbuf *packetTempBuffer;
    bool link = false;
    static unsigned char ucBuffer[ENET_FRAME_MAX_FRAMELEN];
    unsigned char *pucBuffer = ucBuffer;
    unsigned char *pucChar;
    uint32_t phyAddr = 0;

    LWIP_ASSERT("Output packet buffer empty", packetBuffer);

    /* Initiate transfer. */
    if (packetBuffer->len == packetBuffer->tot_len)
    {
        /* No pbuf chain, don't have to copy -> faster. */
        pucBuffer = &((unsigned char *)packetBuffer->payload)[ETH_PAD_SIZE];
    }
    else
    {
        /* pbuf chain, copy into contiguous ucBuffer. */
        if (packetBuffer->tot_len >= ENET_FRAME_MAX_FRAMELEN)
        {
            return ERR_BUF;
        }
        else
        {
            pucChar = ucBuffer;

            for (packetTempBuffer = packetBuffer; packetTempBuffer != NULL; packetTempBuffer = packetTempBuffer->next)
            {
                /* Send the data from the pbuf to the interface, one pbuf at a
                time. The size of the data in each pbuf is kept in the ->len
                variable. */
                /* send data from(packetTempBuffer->payload, packetTempBuffer->len); */
                if (packetTempBuffer == packetBuffer)
                {
                    memcpy(pucChar, &((char *)packetTempBuffer->payload)[ETH_PAD_SIZE],
                           packetTempBuffer->len - ETH_PAD_SIZE);
                    pucChar += packetTempBuffer->len - ETH_PAD_SIZE;
                }
                else
                {
                    memcpy(pucChar, packetTempBuffer->payload, packetTempBuffer->len);
                    pucChar += packetTempBuffer->len;
                }
            }
        }
    }

    /* Send a multicast frame when the PHY is link up. */
    if (kStatus_Success == PHY_GetLinkStatus(ENET, phyAddr, &link))
    {
        if (link)
        {
            if (kStatus_Success == ENET_SendFrame(ENET, &g_handle, pucBuffer, packetBuffer->tot_len - ETH_PAD_SIZE))
            {
                return ERR_OK;
            }
        }
    }
    return ERR_IF;
}

/**
  * @brief Should allocate a pbuf and transfer the bytes of the incoming
  * packet from the interface into the pbuf.
  *
  * @param netif the lwip network interface structure for this ethernetif
  * @return a pbuf filled with the received packet (including MAC header)
  *         NULL on memory error
  */
static struct pbuf * low_level_input(struct netif *netif)
{
  struct pbuf *packetBuffer = NULL, *currentPacket = NULL;
  uint8_t *buffer;
  uint32_t bufferoffset = 0;
  uint32_t payloadoffset = 0;
  uint32_t byteslefttocopy = 0;
	uint32_t length = 0;

  /* get received frame */
	if (kStatus_Success != ENET_GetRxFrameSize(&g_handle, &length))
		return NULL;

  if (length > 0)
  {
    /* We allocate a pbuf chain of pbufs from the Lwip buffer pool */
	  packetBuffer = pbuf_alloc(PBUF_RAW, length, PBUF_POOL);
	  buffer = packetBuffer->payload;
	  ENET_ReadFrame(ENET, &g_handle, buffer, length);
  }

  if (packetBuffer != NULL)
  {
    bufferoffset = 0;

    for(currentPacket = packetBuffer; currentPacket != NULL; currentPacket = currentPacket->next)
    {
      byteslefttocopy = currentPacket->len;
      payloadoffset = 0;

      /* Check if the length of bytes to copy in current pbuf is bigger than Rx buffer size */
      while( (byteslefttocopy + bufferoffset) > ENET_RXBUFF_SIZE )
      {
        /* Copy data to pbuf */
        memcpy( (uint8_t*)((uint8_t*)currentPacket->payload + payloadoffset), (uint8_t*)((uint8_t*)buffer + bufferoffset), (ENET_RXBUFF_SIZE - bufferoffset));

        byteslefttocopy = byteslefttocopy - (ENET_RXBUFF_SIZE - bufferoffset);
        payloadoffset = payloadoffset + (ENET_RXBUFF_SIZE - bufferoffset);
        bufferoffset = 0;
      }

      /* Copy remaining data in pbuf */
      memcpy( (uint8_t*)((uint8_t*)currentPacket->payload + payloadoffset), (uint8_t*)((uint8_t*)buffer + bufferoffset), byteslefttocopy);
      bufferoffset = bufferoffset + byteslefttocopy;
    }
  }
  return packetBuffer;
}

/**
 * This function should be called when a packet is ready to be read
 * from the interface. It uses the function low_level_input() that
 * should handle the actual reception of bytes from the network
 * interface. Then the type of the received packet is determined and
 * the appropriate input function is called.
 *
 * @param netif the lwip network interface structure for this ethernetif
 */
void ethernetif_input(void *param) {
	struct pbuf *p;
	struct netif *netif = (struct netif *) param;

	for (;;) {
		if (xSemaphoreTake( enet_receive_sem, portMAX_DELAY) == pdTRUE) {
			do {
				p = low_level_input(netif);
				if (p != NULL) {
					if (tcpip_input(p, netif) != ERR_OK) {
						pbuf_free(p);
					}
				}
			} while (p != NULL);
		}
	}
}

/**
 * Should be called at the beginning of the program to set up the
 * network interface. It calls the function low_level_init() to do the
 * actual setup of the hardware.
 *
 * This function should be passed as a parameter to netif_add().
 *
 * @param netif the lwip network interface structure for this ethernetif
 * @return ERR_OK if the loopif is initialized
 *         ERR_MEM if private data couldn't be allocated
 *         any other err_t on error
 */
err_t ethernetif_init(struct netif *netif)
{
  LWIP_ASSERT("netif != NULL", (netif != NULL));

#if LWIP_NETIF_HOSTNAME
  /* Initialize interface hostname */
  netif->hostname = "lwip";
#endif /* LWIP_NETIF_HOSTNAME */

  netif->name[0] = IFNAME0;
  netif->name[1] = IFNAME1;

  netif->output = etharp_output;
  netif->linkoutput = low_level_output;

  NVIC_SetPriority(ENET_Receive_IRQn, 0xf);
  NVIC_SetPriority(ENET_Transmit_IRQn, 0xf);
  NVIC_SetPriority(ENET_Error_IRQn, 0xf);
  /* initialize the hardware */
  return low_level_init(netif);
}

/**
  * @brief  This function sets the netif link status.
  * @param  netif: the network interface
  * @retval None
  */
void ethernetif_set_link(void const *argument) {
	bool link = false;
	uint32_t phyAddr = 0, linkStatus = 0;
	status_t error = kStatus_Success;
	struct netif *netif = (struct netif *) argument;

	for (;;) {
		/* Read PHY*/
		error = PHY_GetLinkStatus(ENET, phyAddr, &link);

		/* Check whether the link is up or down*/
		if (error == kStatus_Success) {
			if ((link == true) && (linkStatus == 0)) {
				netif_set_link_up(netif);
				linkStatus = 1;
			} else if ((link == false) && (linkStatus == 1)){
				netif_set_link_down(netif);
				linkStatus = 0;
			}
		}

		/* wait 250 ms */
		TickType_t ticks = INTERFACE_LINK_STATUS_PERIOD / portTICK_PERIOD_MS;
		vTaskDelay(ticks ? ticks : 1);          /* Minimum delay = 1 tick */
	}
}

/**
  * @brief  Link callback function, this function is called on change of link status
  *         to update low level driver configuration.
  * @param  netif: The network interface
  * @retval None
  */
void ethernetif_update_config(struct netif *netif) {
	status_t result = kStatus_Success;

	uint32_t bssReg;
	uint32_t counter = 1000;

	enet_config_t config;
	phy_speed_t speed;
	phy_duplex_t duplex;

	ENET_Type * base = ENET;
	uint32_t phyAddr = 0;

	ENET_GetDefaultConfig(&config);

	if (netif_is_link_up(netif)) {
		/* (Re)start the negotiation */

		/* Reset PHY. */
		result = PHY_Write(base, phyAddr, PHY_BASICCONTROL_REG, PHY_BCTL_RESET_MASK);
		if (result == kStatus_Success)
		{
			/* Set the negotiation. */
			result = PHY_Write(base, phyAddr, PHY_AUTONEG_ADVERTISE_REG,
							   (PHY_100BASETX_FULLDUPLEX_MASK | PHY_100BASETX_HALFDUPLEX_MASK |
								PHY_10BASETX_FULLDUPLEX_MASK | PHY_10BASETX_HALFDUPLEX_MASK | 0x1U));
			if (result == kStatus_Success)
			{
				result = PHY_Write(base, phyAddr, PHY_BASICCONTROL_REG,
								   (PHY_BCTL_AUTONEG_MASK | PHY_BCTL_RESTART_AUTONEG_MASK));
				if (result == kStatus_Success)
				{
					/* Check auto negotiation complete. */
					while (counter --)
					{
						result = PHY_Read(base, phyAddr, PHY_BASICSTATUS_REG, &bssReg);
						if ( result == kStatus_Success)
						{
							if ((bssReg & PHY_BSTATUS_AUTONEGCOMP_MASK) != 0)
							{
								break;
							}
						}

						if (!counter)
						{
							/* Default config */
							config.miiSpeed = kENET_MiiSpeed100M;
							config.miiDuplex = kENET_MiiFullDuplex;
							config.interrupt = kENET_RxFrameInterrupt;
						}

						osDelay(2);
					}
				}
			}
		}

		if (kStatus_Success
				== PHY_GetLinkSpeedDuplex(base, phyAddr, &speed, &duplex)) {
			/* Change the MII speed and duplex for actual link status. */
			config.miiSpeed = (enet_mii_speed_t) speed;
			config.miiDuplex = (enet_mii_duplex_t) duplex;
			config.interrupt = kENET_RxFrameInterrupt;
		} else {
			/* Default config */
			config.miiSpeed = kENET_MiiSpeed100M;
			config.miiDuplex = kENET_MiiFullDuplex;
			config.interrupt = kENET_RxFrameInterrupt;
		}

		enet_buffer_config_t buffCfg = {
		ENET_RXBD_NUM,
		ENET_TXBD_NUM, ENET_BuffSizeAlign(ENET_RXBUFF_SIZE), ENET_BuffSizeAlign(
				ENET_TXBUFF_SIZE), &RxBuffDescrip[0], &TxBuffDescrip[0],
				&RxDataBuff[0][0], &TxDataBuff[0][0], };

		ENET_Init(ENET, &g_handle, &config, &buffCfg, netif->hwaddr,
				CLOCK_GetFreq(kCLOCK_CoreSysClk));
		ENET_SetCallback(&g_handle, ethernet_callback, netif);
		ENET_ActiveRead(ENET);

	}

	ethernetif_notify_conn_changed(netif);
}
