/*
 * C
 *
 * Copyright 2013-2016 IS2T. All rights reserved.
 * For demonstration purpose only.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

/*
*********************************************************************************************************
*                                        INCLUDE FILES
*********************************************************************************************************
*/
#include "pit.h"


/*
*********************************************************************************************************
**                                         EXAMPLE
*
*
* In the example configuration:
*       * The PIT clock has a frequency of 50 MHz.
*       * Timer 1 creates an interrupt every 5.12 ms.
*       * Timer 3 creates a trigger event every 30 ms.
*
* First the PIT module must be activated by using PIT_Enable().
* The 50 MHz clock frequency equates to a clock period of 20 ns. Timer 1 needs to trigger
* every 5.12 ms/20 ns = 256000 cycles and timer 3 every 30 ms/20 ns = 1500000 cycles.
*
* NOTE : The Timer Value trigger is calculated as LDVAL trigger = (period / clock period) - 1
*
* Then set the timer value trigger by using PIT_ChnlSetVal(channel_num, value)
*
* The interrupt for Timer channel_num is enabled by using PIT_ChnlIntEnable(channel_num) and the timer is
* started by using PIT_ChnlTimerEnable(channel_num).
*
* For the timer 1, a handler must be set for the interrupt by using the BSP setter :
*
*  BSP_IntVectSet(__interrupt_ID__, __isr_handler__);
*  BSP_IntEn(__interrupt_ID__);
*
* The following example code matches the described setup:
******
*       PIT_Enable();
*
*       PIT_ChnlSetVal(1, 256000 - 1);
*       PIT_ChnlSetVal(3, 1500000 - 1);
*
*       BSP_IntVectSet(BSP_INT_ID_PIT1, __isr_handler__);
*       BSP_IntEn(BSP_INT_ID_PIT1);
*       PIT_ChnlIntEnable(1);
*       PIT_ChnlIntDisable(3);
*
*       PIT_ChnlTimerEnable(1);
*       PIT_ChnlTimerEnable(3);
*
*
*
*
*********************************************************************************************************
*/

/* Structure of initialize PIT */
pit_config_t pitConfig;

/*
*********************************************************************************************************
*                                       INLINE FUNCTIONS
*********************************************************************************************************
*/
/*
*********************************************************************************************************
** PIT Enable/Disable
*
*
* Description : This function Enables or disables the module clock.
*
* Note :
*       - The PIT must be enabled before any other setup is done.
*       - PIT Module is based on Bus Clock
*********************************************************************************************************
*/
void             PIT_Enable(void)
{
    /*
     * pitConfig.enableRunInDebug = false;
     */
    PIT_GetDefaultConfig(&pitConfig);
    PIT_Init(PIT, &pitConfig);
}

void             PIT_Disable(void)
{
    PIT_Deinit(PIT);
}


/*
*********************************************************************************************************
** PIT Freeze Enable/Disable
*
*
* Description : This function Enables or disables the Freeze mode. it allows the timers to be stopped
* when the device enters debug mode.
*********************************************************************************************************
*/
void             PIT_FreezeEnable(void)
{
    PIT_GetDefaultConfig(&pitConfig);
    pitConfig.enableRunInDebug = false;
    PIT_Init(PIT, &pitConfig);
}

void             PIT_FreezeDisable(void)
{
    PIT_GetDefaultConfig(&pitConfig);
    pitConfig.enableRunInDebug = true;
    PIT_Init(PIT, &pitConfig);
}


/*
*********************************************************************************************************
** Channel timer Enable/Disable
*
*
* Description : This function Enables or disables the timer count for the given channel timer
*********************************************************************************************************
*/
void             PIT_ChnlTimerEnable(pit_chnl_t channel)
{
    PIT_EnableInterrupts(PIT, channel, PIT_TCTRL_TEN_MASK);
}

void             PIT_ChnlTimerDisable(pit_chnl_t channel)
{
    PIT_DisableInterrupts(PIT, channel, PIT_TCTRL_TEN_MASK);
}


/*
*********************************************************************************************************
* Channel Interrupt Enable/Disable
*
*
* Description : This function Enables or disables the interrupt for the given channel timer
*
* Note : When an interrupt is pending, enabling the interrupt will immediately cause an interrupt event.
* To avoid this, the flag must be cleared first.
*********************************************************************************************************
*/
void             PIT_ChnlIntEnable(pit_chnl_t channel)
{
    PIT_EnableInterrupts(PIT, channel, PIT_TCTRL_TIE_MASK);
}

void             PIT_ChnlIntDisable(pit_chnl_t channel)
{
    PIT_DisableInterrupts(PIT, channel, PIT_TCTRL_TIE_MASK);
}


/*
*********************************************************************************************************
** Channel Timer Value Set
*
*
* Description : This function sets the timer start value.
* The timer will count down until it reaches 0, then it will generate an interrupt and load this register
* value again. Setting a new value will not restart the timer, instead the value will be loaded once the
* timer expires. To abort the current cycle and start a timer period with the new value, the timer must
* be disabled and enabled again.
*********************************************************************************************************
*/
void             PIT_ChnlSetVal(pit_chnl_t channel,
                                       uint32_t count)
{
    PIT_SetTimerPeriod(PIT, channel, count);
}


/*
*********************************************************************************************************
* Channel Current Timer Value Get
*
*
* Description : This function returns the current timer value.
*
* Notes :
*      - If the timer is disabled, zero is returned
*      - The timer uses a downcounter.
*      - The timer values are frozen in debug mode if the freeze mode is enabled
*********************************************************************************************************
*/
uint32_t       PIT_ChnlGetCurVal(pit_chnl_t channel)
{
    return PIT_GetCurrentTimerCount(PIT, channel);
}


/*
*********************************************************************************************************
* Channel Interrupt Flag Reset
*
*
* Description : This function resets the timer interrupt flag
*********************************************************************************************************
*/
void             PIT_ChnlFlagRst(pit_chnl_t channel)
{
    PIT->CHANNEL[channel].TFLG |= PIT_TFLG_TIF_MASK;
}
